@@ -1,12 +1,17 @@
 from collections import deque
-from itertools import accumulate
+from itertools import (accumulate,
+                       chain,
+                       groupby,
+                       repeat)
 from typing import (Iterable,
                     List,
                     Optional,
                     Sequence,
                     Set)
 
+from decision.partition import coin_change
 from reprit.base import generate_repr
+from robust import parallelogram
 from robust.angular import (Orientation,
                             orientation)
 from robust.linear import (SegmentsRelationship,
@@ -19,20 +24,31 @@ from sect.hints import (Contour,
 from .contracts import (is_point_inside_circumcircle,
                         points_form_convex_quadrilateral)
 from .events_queue import EventsQueue
-from .subdivisional import (QuadEdge,
-                            edge_to_endpoints,
-                            edge_to_neighbours,
-                            edge_to_non_adjacent_vertices,
-                            edge_to_segment)
+from .quad_edge import (QuadEdge,
+                        edge_to_neighbours,
+                        edge_to_non_adjacent_vertices,
+                        edges_with_opposites)
 from .sweep import sweep
-from .utils import (coin_change,
+from .utils import (ceil_log2,
                     contour_to_segments,
                     flatten,
-                    normalize_triangle,
+                    normalize_contour,
                     pairwise,
+                    to_min_max,
                     to_unique_objects)
 
 
+class CollinearFragmentsKey:
+    __slots__ = 'edge',
+
+    def __init__(self, edge: QuadEdge) -> None:
+        self.edge = edge
+
+    def __eq__(self, other: 'CollinearFragmentsKey') -> bool:
+        return not parallelogram.signed_area(self.edge.start, self.edge.end,
+                                             other.edge.start, other.edge.end)
+
+
 class Triangulation:
     __slots__ = 'left_edge', 'right_edge', '_triangular_holes_vertices'
 
@@ -48,8 +64,9 @@ class Triangulation:
         points = sorted(to_unique_objects(points))
         lengths = coin_change(len(points), _initializers)
         result = [_initialize_triangulation(points[start:stop])
-                  for start, stop in pairwise(accumulate([0] + lengths))]
-        while len(result) > 1:
+                  for start, stop in pairwise(accumulate(chain((0,),
+                                                               lengths)))]
+        for _ in repeat(None, ceil_log2(len(result))):
             parts_to_merge_count = len(result) // 2 * 2
             result = ([result[offset]._merge_with(result[offset + 1])
                        for offset in range(0, parts_to_merge_count, 2)]
@@ -57,7 +74,7 @@ class Triangulation:
         return result[0]
 
     def constrain(self, constraints: Iterable[Segment]) -> None:
-        endpoints = {edge_to_endpoints(edge) for edge in self._to_edges()}
+        endpoints = {edge.endpoints for edge in self.edges()}
         inner_edges = self._to_unique_inner_edges()
         for constraint in constraints:
             constraint_endpoints = frozenset(constraint)
@@ -65,69 +82,89 @@ class Triangulation:
                 continue
             crossings = _detect_crossings(inner_edges, constraint)
             inner_edges.difference_update(crossings)
-            endpoints.difference_update(edge_to_endpoints(edge)
-                                        for edge in crossings)
+            endpoints.difference_update(edge.endpoints for edge in crossings)
             new_edges = _resolve_crossings(crossings, constraint)
             _set_criterion(edge
                            for edge in new_edges
-                           if edge_to_endpoints(edge) != constraint_endpoints)
-            endpoints.update(edge_to_endpoints(edge) for edge in new_edges)
+                           if edge.endpoints != constraint_endpoints)
+            endpoints.update(edge.endpoints for edge in new_edges)
             inner_edges.update(new_edges)
 
     def bound(self, border_segments: Sequence[Segment]) -> None:
         border_endpoints = {frozenset(segment) for segment in border_segments}
-        non_boundary = {edge
-                        for edge in self._to_boundary_edges()
-                        if edge_to_endpoints(edge) not in border_endpoints}
+
+        def unite_fragments(fragments: Iterable[QuadEdge]) -> Segment:
+            return to_min_max(flatten(fragment.segment
+                                      for fragment in fragments))
+
+        non_boundary = set(flatten(
+                edges_with_opposites(fragment
+                                     for fragment in fragments
+                                     if (fragment.endpoints
+                                         not in border_endpoints))
+                for fragments in (
+                    tuple(fragments)
+                    for _, fragments in groupby(self.unique_boundary_edges(),
+                                                key=CollinearFragmentsKey))
+                if (frozenset(unite_fragments(fragments))
+                    not in border_endpoints)))
         while non_boundary:
             edge = non_boundary.pop()
             non_boundary.remove(edge.opposite)
             candidates = edge_to_neighbours(edge)
             self.delete(edge)
-            non_boundary.update(flatten(
-                    (candidate, candidate.opposite)
+            non_boundary.update(edges_with_opposites(
+                    candidate
                     for candidate in candidates
-                    if edge_to_endpoints(candidate) not in border_endpoints))
+                    if candidate.endpoints not in border_endpoints))
 
     def cut(self, holes: Sequence[Contour]) -> None:
         if not holes:
             return
         events_queue = EventsQueue()
-        for edge in self._to_unique_inner_edges():
-            events_queue.register_edge(edge,
-                                       from_test_contour=False)
         for hole in holes:
             for segment in contour_to_segments(hole):
                 events_queue.register_segment(segment,
-                                              from_test_contour=True)
+                                              from_left=True)
+        for edge in self._to_unique_inner_edges():
+            events_queue.register_edge(edge,
+                                       from_left=False)
         hole_segments_endpoints, candidates = set(), []
         for event in sweep(events_queue):
-            if event.from_test_contour:
+            if event.from_left:
                 hole_segments_endpoints.add(frozenset(event.segment))
             elif event.in_intersection:
                 candidates.append(event.edge)
         for edge in candidates:
-            if edge_to_endpoints(edge) not in hole_segments_endpoints:
+            if edge.endpoints not in hole_segments_endpoints:
                 self.delete(edge)
         self._triangular_holes_vertices.update(frozenset(hole)
                                                for hole in holes
                                                if len(hole) == 3)
 
     def triangles(self) -> List[Triangle]:
+        """
+        Returns triangles of the triangulation.
+
+        >>> points = [(0, 0), (0, 1), (1, 0), (1, 1)]
+        >>> triangulation = Triangulation.from_points(points)
+        >>> (triangulation.triangles()
+        ...  == [((0, 1), (1, 0), (1, 1)), ((0, 0), (1, 0), (0, 1))])
+        True
+        """
         return list(self._triangles())
 
     def _triangles(self) -> Iterable[Triangle]:
         visited_vertices = set(self._triangular_holes_vertices)
-        for edge in self._to_edges():
-            if (edge.orientation_with(edge.left_from_start.end)
-                    is Orientation.COUNTERCLOCKWISE):
-                if edge.left_from_start.end != edge.opposite.right_from_start.end:
-                    continue
+        for edge in self.edges():
+            if (edge.left_from_start.end == edge.opposite.right_from_start.end
+                    and (edge.orientation_with(edge.left_from_start.end)
+                         is Orientation.COUNTERCLOCKWISE)):
                 triangle = edge.start, edge.end, edge.left_from_start.end
                 vertices = frozenset(triangle)
                 if vertices not in visited_vertices:
                     visited_vertices.add(vertices)
-                    yield normalize_triangle(triangle)
+                    yield normalize_contour(triangle)
 
     def _merge_with(self, other: 'Triangulation') -> 'Triangulation':
         _merge(self._find_base_edge(other))
@@ -151,22 +188,86 @@ class Triangulation:
         return base_edge
 
     def delete(self, edge: QuadEdge) -> None:
+        """
+        Deletes the edge from the triangulation.
+
+        >>> points = [(0, 0), (0, 1), (1, 0), (1, 1)]
+        >>> triangulation = Triangulation.from_points(points)
+        >>> edges = [triangulation.left_edge, triangulation.right_edge]
+        >>> all(edge in triangulation.edges() for edge in edges)
+        True
+        >>> for edge in edges:
+        ...     triangulation.delete(edge)
+        >>> any(edge in triangulation.edges() for edge in edges)
+        False
+        """
         if edge is self.right_edge or edge.opposite is self.right_edge:
             self.right_edge = self.right_edge.right_from_end.opposite
         if edge is self.left_edge or edge.opposite is self.left_edge:
             self.left_edge = self.left_edge.left_from_start
-        edge.delete()
+        edge.disconnect()
 
-    def _to_boundary_edges(self) -> List[QuadEdge]:
-        return list(flatten((edge, edge.opposite)
-                            for edge in self._to_ccw_boundary_edges()))
+    def boundary_edges(self) -> Iterable[QuadEdge]:
+        """
+        Returns boundary edges of the triangulation in counterclockwise order.
 
-    def _to_edges(self) -> Set[QuadEdge]:
-        for edge in self._to_unique_edges():
+        >>> points = [(0, 0), (0, 1), (1, 0), (1, 1)]
+        >>> triangulation = Triangulation.from_points(points)
+        >>> ([edge.segment for edge in triangulation.boundary_edges()]
+        ...  == [((0, 0), (1, 0)), ((1, 0), (0, 0)),
+        ...      ((1, 0), (1, 1)), ((1, 1), (1, 0)),
+        ...      ((1, 1), (0, 1)), ((0, 1), (1, 1)),
+        ...      ((0, 1), (0, 0)), ((0, 0), (0, 1))])
+        True
+        """
+        return edges_with_opposites(self.unique_boundary_edges())
+
+    def edges(self) -> Iterable[QuadEdge]:
+        """
+        Returns edges of the triangulation.
+
+        >>> points = [(0, 0), (0, 1), (1, 0), (1, 1)]
+        >>> triangulation = Triangulation.from_points(points)
+        >>> ([edge.segment for edge in triangulation.edges()]
+        ...  == [((1, 1), (1, 0)), ((1, 0), (1, 1)),
+        ...      ((1, 0), (0, 1)), ((0, 1), (1, 0)),
+        ...      ((0, 1), (1, 1)), ((1, 1), (0, 1)),
+        ...      ((0, 1), (0, 0)), ((0, 0), (0, 1)),
+        ...      ((0, 0), (1, 0)), ((1, 0), (0, 0))])
+        True
+        """
+        return edges_with_opposites(self.unique_edges())
+
+    def unique_boundary_edges(self) -> Iterable[QuadEdge]:
+        """
+        Returns boundary edges of the triangulation in counterclockwise order.
+
+        >>> points = [(0, 0), (0, 1), (1, 0), (1, 1)]
+        >>> triangulation = Triangulation.from_points(points)
+        >>> ([edge.segment for edge in triangulation.unique_boundary_edges()]
+        ...  == [((0, 0), (1, 0)), ((1, 0), (1, 1)),
+        ...      ((1, 1), (0, 1)), ((0, 1), (0, 0))])
+        True
+        """
+        start = self.left_edge
+        edge = start
+        while True:
             yield edge
-            yield edge.opposite
+            if edge.right_from_end is start:
+                break
+            edge = edge.right_from_end
 
-    def _to_unique_edges(self) -> Set[QuadEdge]:
+    def unique_edges(self) -> Iterable[QuadEdge]:
+        """
+        Returns edges of the triangulation with unique endpoints.
+
+        >>> points = [(0, 0), (0, 1), (1, 0), (1, 1)]
+        >>> triangulation = Triangulation.from_points(points)
+        >>> ([edge.segment for edge in triangulation.unique_edges()]
+        ...  == [((1, 1), (1, 0)), ((1, 0), (0, 1)), ((0, 1), (1, 1)),
+        ...      ((0, 1), (0, 0)), ((0, 0), (1, 0))])
+        True
+        """
         visited_edges = set()
         is_visited, visit_multiple = (visited_edges.__contains__,
                                       visited_edges.update)
@@ -180,21 +281,8 @@ class Triangulation:
             queue.extend((edge.left_from_start, edge.left_from_end,
                           edge.right_from_start, edge.right_from_end))
 
-    def _to_ccw_boundary_edges(self) -> Iterable[QuadEdge]:
-        start = self.left_edge
-        edge = start
-        while True:
-            yield edge
-            if edge.right_from_end is start:
-                break
-            edge = edge.right_from_end
-
-    def _to_inner_edges(self) -> Set[QuadEdge]:
-        return set(self._to_edges()).difference(self._to_boundary_edges())
-
     def _to_unique_inner_edges(self) -> Set[QuadEdge]:
-        return (set(self._to_unique_edges())
-                .difference(self._to_boundary_edges()))
+        return set(self.unique_edges()).difference(self.boundary_edges())
 
 
 def _resolve_crossings(crossings: List[QuadEdge],
@@ -210,7 +298,7 @@ def _resolve_crossings(crossings: List[QuadEdge],
                                              first_non_edge_vertex,
                                              second_non_edge_vertex)):
             edge.swap()
-            if (segments_relationship(edge_to_segment(edge), constraint)
+            if (segments_relationship(edge.segment, constraint)
                     is SegmentsRelationship.CROSS):
                 crossings.append(edge)
             else:
@@ -224,8 +312,8 @@ def _detect_crossings(inner_edges: Iterable[QuadEdge],
                       constraint: Segment) -> List[QuadEdge]:
     return [edge
             for edge in inner_edges
-            if segments_relationship(edge_to_segment(edge), constraint)
-            is SegmentsRelationship.CROSS]
+            if (segments_relationship(edge.segment, constraint)
+                is SegmentsRelationship.CROSS)]
 
 
 def _merge(base_edge: QuadEdge) -> None:
@@ -254,7 +342,7 @@ def _to_left_candidate(base_edge: QuadEdge) -> Optional[QuadEdge]:
            and (base_edge.orientation_with(result.left_from_start.end)
                 is Orientation.CLOCKWISE)):
         next_candidate = result.left_from_start
-        result.delete()
+        result.disconnect()
         result = next_candidate
     return result
 
@@ -269,7 +357,7 @@ def _to_right_candidate(base_edge: QuadEdge) -> Optional[QuadEdge]:
            and (base_edge.orientation_with(result.right_from_start.end)
                 is Orientation.CLOCKWISE)):
         next_candidate = result.right_from_start
-        result.delete()
+        result.disconnect()
         result = next_candidate
     return result
 
