@@ -11,7 +11,7 @@ from .quad_edge import QuadEdge
 
 
 class EventsQueueKey:
-    __slots__ = ('event',)
+    __slots__ = 'event',
 
     def __init__(self, event: Event) -> None:
         self.event = event
@@ -45,11 +45,11 @@ class EventsQueueKey:
                 return other_end_orientation is (Orientation.COUNTERCLOCKWISE
                                                  if event.is_left_endpoint
                                                  else Orientation.CLOCKWISE)
-            elif event.from_test_contour is not other_event.from_test_contour:
-                return other_event.from_test_contour
+            elif event.from_left is not other_event.from_left:
+                return other_event.from_left
             else:
-                end_x, end_y = event.end
-                other_end_x, other_end_y = other_event.end
+                _, end_y = event.end
+                _, other_end_y = other_event.end
                 return end_y < other_end_y
 
 
@@ -67,37 +67,32 @@ class EventsQueue:
 
     def register_segment(self, segment: Segment,
                          *,
-                         from_test_contour: bool) -> None:
+                         from_left: bool) -> None:
         start, end = sorted(segment)
-        start_event = Event(True, start, None, from_test_contour,
-                            EdgeKind.NORMAL)
-        end_event = Event(False, end, start_event, from_test_contour,
-                          EdgeKind.NORMAL)
+        start_event = Event(True, start, None, from_left, EdgeKind.NORMAL)
+        end_event = Event(False, end, start_event, from_left, EdgeKind.NORMAL)
         start_event.complement = end_event
         self._queue.push(start_event)
         self._queue.push(end_event)
 
     def register_edge(self, edge: QuadEdge,
                       *,
-                      from_test_contour: bool) -> None:
+                      from_left: bool) -> None:
         start, end = sorted((edge.start, edge.end))
-        start_event = Event(True, start, None, from_test_contour,
-                            EdgeKind.NORMAL, edge)
-        end_event = Event(False, end, start_event, from_test_contour,
-                          EdgeKind.NORMAL, edge)
+        start_event = Event(True, start, None, from_left, EdgeKind.NORMAL,
+                            edge)
+        end_event = Event(False, end, start_event, from_left, EdgeKind.NORMAL,
+                          edge)
         start_event.complement = end_event
         self._queue.push(start_event)
         self._queue.push(end_event)
 
     def divide_segment(self, event: Event, point: Point) -> None:
         left_event = Event(True, point, event.complement,
-                           event.from_test_contour, EdgeKind.NORMAL,
+                           event.from_left, EdgeKind.NORMAL,
                            event.edge)
-        right_event = Event(False, point, event, event.from_test_contour,
+        right_event = Event(False, point, event, event.from_left,
                             EdgeKind.NORMAL, event.edge)
         event.complement.complement, event.complement = left_event, right_event
         self._queue.push(left_event)
         self._queue.push(right_event)
-
-    def clear(self) -> None:
-        self._queue.clear()
